/**
 * MULe/2010L: JPEG compression - DCT settings (computer exercise)
 */
#include <iostream>
#define _USE_MATH_DEFINES
#include <cmath>
#include <cv.h>
#include <highgui.h>

using namespace std;
using namespace cv;

/**
 * block size is 8x8 pixels
 */
static const int N = 8;

/**
 * auxiliary function
 */
static double lambda (int k)
{
	if (0 == k)
		return 1.0/sqrt(2.0);
	else
		return 1.0;
}

/**
 * forward 1D DCT
 */
static void fdct1D (const double *f, double *c)
{
	for (int k=0; k<N; k++)
	{
		c[k] = 0.0;
		for (int n=0; n<N; n++)
		{
			c[k] += f[n] * lambda(k) * sqrt(2.0/N) * cos(k*M_PI/N*(n+0.5));
		}
	}
}

/**
 * inverse 1D DCT
 */
static void idct1D (const double *c, double *f)
{
	for (int n=0; n<N; n++)
	{
		f[n] = 0.0;
		for (int k=0; k<N; k++)
		{
			f[n] += c[k] * lambda(k) * sqrt(2.0/N) * cos(k*M_PI/N*(n+0.5));
		}
	}
}

/**
 * forward 2D DCT
 */
static void fdct2D (const Mat& src, Mat& dst)
{
#if 1
	// OpenCV
	dct (src, dst);
#endif

#if 0
	// series of 1D DCT on the rows and columns
	for (int y=0; y<N; y++)
		fdct1D( &src.at<double>(y,0), &dst.at<double>(y,0));
	Mat tmp = dst.t();
	for (int y=0; y<N; y++)
		fdct1D( &tmp.at<double>(y,0), &dst.at<double>(y,0));
	dst = dst.t();
#endif

#if 0
	// 2D DCT directly
	for (int k=0; k<N; k++)
		for (int j=0; j<N; j++)
		{
			double & c = dst.at<double>(k,j);
			c = 0;
			for (int n=0; n<N; n++)
				for (int m=0; m<N; m++)
					c += src.at<double>(n,m) * lambda(k) * lambda(j) * 2/N * cos(k*M_PI/N*(n+0.5)) * cos(j*M_PI/N*(m+0.5));
		}
#endif
}

/**
 * inverse 2D DCT
 */
static void idct2D (const Mat& src, Mat& dst)
{
#if 1
	// OpenCV
	dct (src, dst, DCT_INVERSE);
#endif

#if 0
	// series of 1D inverse DCT on the rows and columns
	for (int y=0; y<N; y++)
		idct1D( &src.at<double>(y,0), &dst.at<double>(y,0));
	Mat tmp = dst.t();
	for (int y=0; y<N; y++)
		idct1D( &tmp.at<double>(y,0), &dst.at<double>(y,0));
	dst = dst.t();
#endif

#if 0
	// 2D inverse DCT directly
	for (int n=0; n<N; n++)
		for (int m=0; m<N; m++)
		{
			double & f = dst.at<double>(n,m);
			f = 0;
			for (int k=0; k<N; k++)
				for (int j=0; j<N; j++)
					f += src.at<double>(k,j) * lambda(k) * lambda(j) * 2/N * cos(k*M_PI/N*(n+0.5)) * cos(j*M_PI/N*(m+0.5));
		}
#endif
}

/**
 * coefficient quantization
 */
static void quantize (const Mat& src, Mat& dst)
{
	src.copyTo (dst);

#if 0
	// simple coefficients zeroing
	for (int y=0; y<N; y++)
		for (int x=0; x<N; x++)
			if (x > 1 || y > 1)
				dst.at<double> (y,x) = 0.0;
#endif

#if 0
	// as it JPEG does
	Mat Q = (Mat_<double>(8,8) <<
		16, 11, 10, 16, 124, 140, 151, 161,
		12, 12, 14, 19, 126, 158, 160, 155,
		14, 13, 16, 24, 140, 157, 169, 156,
		14, 17, 22, 29, 151, 187, 180, 162,
		18, 22, 37, 56, 168, 109, 103, 177,
		24, 35, 55, 64, 181, 104, 113, 192,
		49, 64, 78, 87, 103, 121, 120, 101,
		72, 92, 95, 98, 112, 100, 103, 199);

	int quality = 5; // 1..100

	Q *= ( quality <= 50 ? 50/quality : 2-100/quality );

	dst = Mat_<double>(Mat_<int>(Mat(src/Q))).mul(Q);
#endif
}

/**
 * demonstration of image compression using DCT
 */
static void demo (const Mat &input)
{
	// here are the DCT coefficients stored
	Mat coeff (input.rows, input.cols, DataType<double>::type);

	// coefficients after quantization
	Mat quant (input.rows, input.cols, DataType<double>::type);

	// decoded image
	Mat decod (input.rows, input.cols, DataType<double>::type);

	// the image is processed in blocks
	for (int y = 0; y < input.rows; y+=N)
		for (int x = 0; x < input.cols; x+=N)
		{
			// this rectangle becomes currently processed block
			Rect rect (x, y, N, N);

			// auxiliary variables that represent corresponding block
			Mat inputBlock (input, rect);
			Mat coeffBlock (coeff, rect);
			Mat quantBlock (quant, rect);
			Mat decodBlock (decod, rect);

			// at first, image block is transformed into coefficients by the DCT
			fdct2D (inputBlock, coeffBlock);

			// afterwards, quantization of these coefficients is performed (lossy compression)
			quantize (coeffBlock, quantBlock);

			// the original image block is reconstructed from the quantized coefficient
			idct2D (quantBlock, decodBlock);
		}

	// show the whole process
	imshow ("input",      Mat_<uchar> (input) );
	imshow ("transform",  abs (coeff/15) );
	imshow ("quantized",  abs (quant/15) );
	imshow ("decoded",    Mat_<uchar> (decod) );
	imshow ("difference", Mat_<uchar> ( Mat (abs (input-decod)) ) );
}

/**
 * program entry
 */
int main (int argc, char **argv)
{
	const char *imagename = argc > 1
		? argv[1] : "Lenna.png";

	cout << "Loading " << imagename << endl;

	Mat img = Mat_<double> ( imread (imagename, CV_LOAD_IMAGE_GRAYSCALE) );
	if (!img.data)
		return -1;

	demo (img);

	waitKey ();

	return 0;
}
